/* My first OpenGL program.
 *
 * Makes use of Mark Kilgard's GLUT, but there are some Windows-specific
 * code, namely the use of POINT, GetCursorPos(), SetCursorPos(),
 * GetAsyncKeyState(), and PlaySound().
 */


#include <GL/glut.h>
#include <stdio.h>
#include "more_math.h"
#include "misc.h"
#include "world.h"
#include "timer.h"

/**********
* Defines
**********/
#define DEFAULT_WIDTH	300	// Default width of window.
#define DEFAULT_HEIGHT	300	// Default height of window.


/**********
* Globals
**********/
// Directional light source properties.
GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
GLfloat light_position[] = { 0.0, 0.0, 0.0, 1.0 };

// Scene objects
world_t *world;


/**********
* Function prototypes
**********/
void toggle_fullscreen();
void entry(int state);
void mouse(int button, int state, int x, int y);
//void motion(int x, int y);
//void passive_motion(int x, int y);
void special_keys(int key, int x, int y);
void idle();
void redraw();
void reshape(int w, int h);
void init();
void cleanup();


/**********
* Toggles fullscreen mode.
**********/
void toggle_fullscreen() {
	static char is_fullscreen = 0;
	static int width = glutGet(GLUT_WINDOW_WIDTH);
	static int height = glutGet(GLUT_WINDOW_HEIGHT);

	if (is_fullscreen) {
		glutReshapeWindow(width, height);
		is_fullscreen = 0;
	}
	else {
		width = glutGet(GLUT_WINDOW_WIDTH);
		height = glutGet(GLUT_WINDOW_HEIGHT);
		glutFullScreen();
		is_fullscreen = 1;
	}
}


/**********
* Called whenever mouse enters/leaves the window.
**********/
void entry(int state)
{
	if (state == GLUT_LEFT)
		world->suspend();
	else {
		SetCursorPos(glutGet(GLUT_WINDOW_X)+MOUSE_MARGIN, glutGet(GLUT_WINDOW_Y)+MOUSE_MARGIN);
		world->resume();
	}
}


/**********
* Called whenever state of a mouse button changes.
**********/
void mouse(int button, int state, int x, int y)
{
	if ( (button == GLUT_RIGHT_BUTTON) && (state == GLUT_DOWN) ) {
		world->reset();
		glutPostRedisplay();
	}
}


/**********
* Rotate object based on movement of mouse. Called whenever mouse moves
* while left button is down.
**********/
/*void motion(int x, int y)
{
	obj->yaw_left( (x - old_mouse_x) * MOUSE_SENSITIVITY );
//	obj->change_yaw( (x - old_mouse_x) * MOUSE_SENSITIVITY );
	glutPostRedisplay();
	old_mouse_x = x;
	old_mouse_y = y;

	// If left button used for firing, call passive_motion(x,y) so camera
	// continues to move. Also remove the above old_mouse_? = ?.
//	passive_motion(x, y);
}*/


/**********
* Ordinary keyboard callback.
**********/
void keyboard(unsigned char key, int x, int y)
{
	switch (key) {
	case 27: // ESC
		if (world->is_suspended()) {
			glutSetCursor(GLUT_CURSOR_NONE);
			world->resume();
		}
		else {
			glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
			world->suspend();
		}
		break;

	default:
		break;
	}
}


/**********
* Called whenever a special keyboard event occurs.
**********/
void special_keys(int key, int x, int y)
{
	// ALT-F4 exits.
	if ( (key == GLUT_KEY_F4) && (glutGetModifiers() == GLUT_ACTIVE_ALT) )
		cleanup();

	// F1 toggles fullscreen.
	if ( key == GLUT_KEY_F1 )
		toggle_fullscreen();
}


/**********
* Called whenever other messages being processed.
**********/
void idle()
{
	update_timer();
	world->process_input();	// Polling is used in order to obtain better response.
	world->update();
	glutPostRedisplay();
}


/**********
* Called whenever window needs to be redrawn.
**********/
void redraw()
{
	char fps[15];

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

//	update_timer();

	// Move light with scene (i.e. light is stationary relative to the
	// scene instead of moving with the viewer).
//	glLightfv(GL_LIGHT0, GL_POSITION, light_position);

	// Draw axes.
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_LINES);
		glNormal3f(0.0f,1.0f,0.0f);
		glVertex3f(-20.0f,0.0f,0.0f);
		glVertex3f(20.0f,0.0f,0.0f);

		glNormal3f(0.0f,1.0f,0.0f);
		glVertex3f(0.0f,-20.0f,0.0f);
		glVertex3f(0.0f,20.0f,0.0f);

		glNormal3f(0.0f,1.0f,0.0f);
		glVertex3f(0.0f,0.0f,-20.0f);
		glVertex3f(0.0f,0.0f,20.0f);
	glEnd();

	world->draw();

	// Output framerate.
	sprintf(fps, "fps: %.2f", get_fps());
	text(30, 30, 12.0, "%s", fps);

	glutSwapBuffers();
}


/**********
* Called whenever window is resized.
**********/
void reshape(int w, int h)
{
	glViewport( 0,0, (GLsizei)w, (GLsizei)h );
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 0.5, 75.0);
	glMatrixMode(GL_MODELVIEW);
//	if ( camera->is_enabled() )
//		camera->refresh();
}


/**********
* Initialize environment.
**********/
void init()
{
	// Setup a light source.
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	// Depth testing, enable/disable GL states, etc.
	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);
	glMateriali(GL_FRONT, GL_SHININESS, 100);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
//	glDisable(GL_DITHER);
	glShadeModel(GL_SMOOTH);
	glutSetCursor(GLUT_CURSOR_NONE);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	// Default to fullscreen.
	toggle_fullscreen();

	// Create world.
	world = new world_t;

	// Play startup sound. Windows specific code here.
	PlaySound("sounds/start.wav", NULL, SND_FILENAME | SND_ASYNC);
}


/**********
* Call this to exit program.
**********/
void cleanup()
{
	delete world;

	exit(0);
}

/**********
* Main
**********/
int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(DEFAULT_WIDTH, DEFAULT_HEIGHT);
//	glutInitWindowSize(glutGet(GLUT_SCREEN_WIDTH), glutGet(GLUT_SCREEN_HEIGHT));
	glutCreateWindow("Cheeze v0.0.3");

	glutEntryFunc(entry);
	glutMouseFunc(mouse);
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(special_keys);
	glutIdleFunc(idle);
	glutDisplayFunc(redraw);
	glutReshapeFunc(reshape);

	init();
	glutMainLoop();
	return 0;
}