#include "more_math.h"


/* Lifted from the Mesa 3-D graphics library.	-Cliff Lee
 *
 * Compute the inverse of a 4x4 matrix.
 *
 * From an algorithm by V. Strassen, 1969, _Numerishe Mathematik_, vol. 13,
 * pp. 354-356.
 * 60 multiplies, 24 additions, 10 subtractions, 8 negations, 2 divisions,
 * 48 assignments, _0_ branches
 *
 * This implementation by Scott McCaskill
 */
 
enum {
    M00 = 0, M01 = 4, M02 = 8, M03 = 12,
    M10 = 1, M11 = 5, M12 = 9, M13 = 13,
    M20 = 2, M21 = 6, M22 = 10,M23 = 14,
    M30 = 3, M31 = 7, M32 = 11,M33 = 15
};

void invert_matrix( const GLfloat *m, GLfloat *out )
{
   GLfloat r1[2][2], r2[2][2], r3[2][2], r4[2][2], r5[2][2], r6[2][2],
         r7[2][2];
   const GLfloat * A = m;
   GLfloat *       C = out;
   GLfloat one_over_det;

   /*
    * A is the 4x4 source matrix (to be inverted).
    * C is the 4x4 destination matrix
    * a11 is the 2x2 matrix in the upper left quadrant of A
    * a12 is the 2x2 matrix in the upper right quadrant of A
    * a21 is the 2x2 matrix in the lower left quadrant of A
    * a22 is the 2x2 matrix in the lower right quadrant of A
    * similarly, cXX are the 2x2 quadrants of the destination matrix
    */

   /* R1 = inverse( a11 ) */
   one_over_det = 1.0f / ( ( A[M00] * A[M11] ) - ( A[M10] * A[M01] ) );
   r1[0][0] = one_over_det * A[M11];
   r1[0][1] = one_over_det * -A[M01];
   r1[1][0] = one_over_det * -A[M10];
   r1[1][1] = one_over_det * A[M00];

   /* R2 = a21 x R1 */
   r2[0][0] = A[M20] * r1[0][0] + A[M21] * r1[1][0];
   r2[0][1] = A[M20] * r1[0][1] + A[M21] * r1[1][1];
   r2[1][0] = A[M30] * r1[0][0] + A[M31] * r1[1][0];
   r2[1][1] = A[M30] * r1[0][1] + A[M31] * r1[1][1];

   /* R3 = R1 x a12 */
   r3[0][0] = r1[0][0] * A[M02] + r1[0][1] * A[M12];
   r3[0][1] = r1[0][0] * A[M03] + r1[0][1] * A[M13];
   r3[1][0] = r1[1][0] * A[M02] + r1[1][1] * A[M12];
   r3[1][1] = r1[1][0] * A[M03] + r1[1][1] * A[M13];

   /* R4 = a21 x R3 */
   r4[0][0] = A[M20] * r3[0][0] + A[M21] * r3[1][0];
   r4[0][1] = A[M20] * r3[0][1] + A[M21] * r3[1][1];
   r4[1][0] = A[M30] * r3[0][0] + A[M31] * r3[1][0];
   r4[1][1] = A[M30] * r3[0][1] + A[M31] * r3[1][1];

   /* R5 = R4 - a22 */
   r5[0][0] = r4[0][0] - A[M22];
   r5[0][1] = r4[0][1] - A[M23];
   r5[1][0] = r4[1][0] - A[M32];
   r5[1][1] = r4[1][1] - A[M33];

   /* R6 = inverse( R5 ) */
   one_over_det = 1.0f / ( ( r5[0][0] * r5[1][1] ) - ( r5[1][0] * r5[0][1] ) );
   r6[0][0] = one_over_det * r5[1][1];
   r6[0][1] = one_over_det * -r5[0][1];
   r6[1][0] = one_over_det * -r5[1][0];
   r6[1][1] = one_over_det * r5[0][0];

   /* c12 = R3 x R6 */
   C[M02] = r3[0][0] * r6[0][0] + r3[0][1] * r6[1][0];
   C[M03] = r3[0][0] * r6[0][1] + r3[0][1] * r6[1][1];
   C[M12] = r3[1][0] * r6[0][0] + r3[1][1] * r6[1][0];
   C[M13] = r3[1][0] * r6[0][1] + r3[1][1] * r6[1][1];

   /* c21 = R6 x R2 */
   C[M20] = r6[0][0] * r2[0][0] + r6[0][1] * r2[1][0];
   C[M21] = r6[0][0] * r2[0][1] + r6[0][1] * r2[1][1];
   C[M30] = r6[1][0] * r2[0][0] + r6[1][1] * r2[1][0];
   C[M31] = r6[1][0] * r2[0][1] + r6[1][1] * r2[1][1];

   /* R7 = R3 x c21 */
   r7[0][0] = r3[0][0] * C[M20] + r3[0][1] * C[M30];
   r7[0][1] = r3[0][0] * C[M21] + r3[0][1] * C[M31];
   r7[1][0] = r3[1][0] * C[M20] + r3[1][1] * C[M30];
   r7[1][1] = r3[1][0] * C[M21] + r3[1][1] * C[M31];

   /* c11 = R1 - R7 */
   C[M00] = r1[0][0] - r7[0][0];
   C[M01] = r1[0][1] - r7[0][1];
   C[M10] = r1[1][0] - r7[1][0];
   C[M11] = r1[1][1] - r7[1][1];

   /* c22 = -R6 */
   C[M22] = -r6[0][0];
   C[M23] = -r6[0][1];
   C[M32] = -r6[1][0];
   C[M33] = -r6[1][1];
}


/**********
* Find the cross product of two vectors.
*
* out
*	3-element array to hold the resulting cross product.
*
* v1, v2
*	3-element arrays holding vectors to cross.
**********/
void cross_product(GLfloat *out, GLfloat *v1, GLfloat *v2)
{
	out[0] = v1[1]*v2[2] - v1[2]*v2[1];
	out[1] = v1[2]*v2[0] - v1[0]*v2[2];
	out[2] = v1[0]*v2[1] - v1[1]*v2[0];
}


/**********
* Convert given vector to a unit vector.
*
* vect
*	3-element array representing vector to be normalized.
**********/
void normalize_vector(GLfloat *vect)
{
	GLfloat length = vector_length(vect);

	vect[0] /= length;
	vect[1] /= length;
	vect[2] /= length;
}


/**********
* Find the length of a vector.
*
* vect
*	3-element array representing the vector whose length is to be found.
**********/
GLfloat vector_length(GLfloat *vect)
{
	return (GLfloat)sqrt(vect[0]*vect[0] + vect[1]*vect[1] + vect[2]*vect[2]);
}


/**********
* Lookup sine from table.
**********/
/*GLfloat fast_sin(GLfloat angle)
{
// if angle not an integer, call sin(angle).

//	angle %= 360;
	if (angle < 180.0f)
		return sin_table[(GLint)angle];
	else
		return -sin_table[(GLint)angle];
}*/