#include "scene_object.h"
#include "timer.h"


/**********
* Gets scaling factor between frames, used in order to adjust speed (slow down
* on fast machines, speed up on slow machines).
*
* Return Value
*	Time between frames, in seconds.
**********/
GLfloat scene_object_t::scale_factor(void)
{
/*	GLint frame_start = get_time();
	GLint frame_end;
	GLfloat ret;

	do
		frame_end = get_time();
	while (frame_end == frame_start);
	ret = (GLfloat)(frame_end - frame_start) / 1000.0;
	frame_start = frame_end;
	return (GLfloat)ret;*/
	return (GLfloat)get_elapsed_time() / 1000.0;
}


/**********
* Move back (i.e. in direction of local positive z-axis), based on
* object's speed.
**********/
void scene_object_t::move_back(void)
{
	GLfloat dist;

	dist = move_speed[2] * scale_factor();
	glPushMatrix();
	glLoadMatrixf(transformation);
	if (!use_inertial)
		glTranslatef(0, 0, dist);
	else {
		glRotatef(-front_angle, 1, 0, 0);
		glTranslatef(0, 0, dist);
		glRotatef(front_angle, 1, 0, 0);
	}
	glGetFloatv(GL_MODELVIEW_MATRIX, transformation);
	glPopMatrix();
}


/**********
* Move right (i.e. in direction of local positive x-axis), based on
* object's speed.
**********/
void scene_object_t::move_right(void)
{
	GLfloat dist;

	dist = move_speed[0] * scale_factor();
	glPushMatrix();
	glLoadMatrixf(transformation);
	if (!use_inertial)
		glTranslatef(dist, 0, 0);
	else {
		glRotatef(-front_angle, 1, 0, 0);
		glTranslatef(dist, 0, 0);
		glRotatef(front_angle, 1, 0, 0);
	}
	glGetFloatv(GL_MODELVIEW_MATRIX, transformation);
	glPopMatrix();
}


/**********
* Move up (i.e. in direction of local positive y-axis), based on
* object's speed.
**********/
void scene_object_t::move_up(void)
{
	GLfloat dist;

	dist = move_speed[1] * scale_factor();
	glPushMatrix();
	glLoadMatrixf(transformation);
	if (!use_inertial)
		glTranslatef(0, dist, 0);
	else {
		glRotatef(-front_angle, 1, 0, 0);
		glTranslatef(0, dist, 0);
		glRotatef(front_angle, 1, 0, 0);
	}
	glGetFloatv(GL_MODELVIEW_MATRIX, transformation);
	glPopMatrix();
}


/**********
* Rotate object around x-axis in positive direction.
*
* angle
*	Number of degrees to rotate.
**********/
void scene_object_t::pitch_back(GLfloat angle)
{
	while (angle >= 360) angle -= 360;
	while (angle <= -360) angle += 360;
	glPushMatrix();
	glLoadMatrixf(transformation);
	glRotatef(angle, 1, 0, 0);
	glGetFloatv(GL_MODELVIEW_MATRIX, transformation);
	glPopMatrix();
	if (use_inertial) front_angle += angle;
}


/**********
* Rotate object around x-axis in positive direction, based on object's
* speed.
**********/
void scene_object_t::pitch_back(void)
{
	GLfloat angle;

	angle = rotate_speed[0] * scale_factor();
	pitch_back(angle);
}


/**********
* Rotate object around y-axis in positive direction.
*
* angle
*	Number of degrees to rotate.
**********/
void scene_object_t::yaw_left(GLfloat angle)
{
	while (angle >= 360) angle -= 360;
	while (angle <= -360) angle += 360;
	glPushMatrix();
	glLoadMatrixf(transformation);
	if (!use_inertial)
		glRotatef(angle, 0, 1, 0);
	else {
		glRotatef(-front_angle, 1, 0, 0);
		glRotatef(angle, 0, 1, 0);
		glRotatef(front_angle, 1, 0, 0);
	}
	glGetFloatv(GL_MODELVIEW_MATRIX, transformation);
	glPopMatrix();

}


/**********
* Rotate object around y-axis in positive direction, based on object's
* speed.
**********/
void scene_object_t::yaw_left(void)
{
	GLfloat angle;

	angle = rotate_speed[1] * scale_factor();
	yaw_left(angle);
}


/**********
* Rotate object around z-axis in positive direction.
*
* angle
*	Number of degrees to rotate.
**********/
void scene_object_t::roll_left(GLfloat angle)
{
	while (angle >= 360) angle -= 360;
	while (angle <= -360) angle += 360;
	glPushMatrix();
	glLoadMatrixf(transformation);
	if (!use_inertial)
		glRotatef(angle, 0, 0, 1);
	else {
		glRotatef(-front_angle, 1, 0, 0);
		glRotatef(angle, 0, 0, 1);
		glRotatef(front_angle, 1, 0, 0);
	}	
	glGetFloatv(GL_MODELVIEW_MATRIX, transformation);
	glPopMatrix();
}


/**********
* Rotate object around z-axis in positive direction, based on object's
* speed.
**********/
void scene_object_t::roll_left(void)
{
	GLfloat angle;

	angle = rotate_speed[2] * scale_factor();
	roll_left(angle);
}


/**********
* Retrieve requested object attribute of integer type. Requesting an
* unknown attribute will return a negative value.
*
* which_attr
*	Name of attribute to retrieve.
*
* Return Value
*	The requested integer attribute. 
**********/
GLint scene_object_t::get_attr_i(GLenum which_attr)
{
	switch (which_attr) {
	case OBJ_DURABILITY:
		return durability;
		break;
	case OBJ_MAX_DURABILITY:
		return max_durability;
		break;
	case OBJ_USE_INERTIAL:
		return use_inertial;
		break;
	default:
		return -1;
		break;
	}
}


/**********
* Retrieve requested object attribute of float type. Requesting an
* unknown attribute will return a negative value.
*
* which_attr
*	Name of attribute to retrieve.
*
* Return Value
*	The requested float attribute. 
**********/
GLfloat scene_object_t::get_attr_f(GLenum which_attr)
{
	switch (which_attr) {
	case OBJ_SPEED_X:
		return move_speed[0];
		break;
	case OBJ_SPEED_Y:
		return move_speed[1];
		break;
	case OBJ_SPEED_Z:
		return move_speed[2];
		break;
	case OBJ_SPEED_PITCH:
		return rotate_speed[0];
		break;
	case OBJ_SPEED_YAW:
		return rotate_speed[1];
		break;
	case OBJ_SPEED_ROLL:
		return rotate_speed[2];
		break;
	case OBJ_MAX_MOVE_SPEED:
		return max_move_speed;
		break;
	case OBJ_MAX_ROTATE_SPEED:
		return max_rotate_speed;
		break;
	case OBJ_FRONT_ANGLE:
		return front_angle;
		break;
	default:
		return -1.0;
		break;
	}
}


/**********
* Retrieve requested object attribute. Requesting an unknown attribute
* will result in no change to attr.
*
* which_attr
*	Name of attribute to retrieve.
*
* attr
*	The requested attribute. The attribute specified by which_attr is
*	expected to be a vector or a set of coordinates, so attr must be a
*	3-element array to contain the returned attribute.
**********/
void scene_object_t::get_attr_fv(GLenum which_attr, GLfloat *attr)
{
	GLfloat m[16];

	switch (which_attr) {
	case OBJ_FRONT:
		invert_matrix(transformation, m);
		attr[0] = m[8];
		attr[1] = m[9];
		attr[2] = m[10];
		break;
	default:
		break;
	}
}


/**********
* Set requested object attribute. Specifying an unknown attribute will
* result in no changes.
*
* which_attr
*	Name of attribute to set.
*
* attr
*	The new attribute value.
**********/
void scene_object_t::set_attr_i(GLenum which_attr, GLint attr)
{
	switch (which_attr) {
	case OBJ_DURABILITY:
		break;
	case OBJ_USE_INERTIAL:
		break;
	default:
		break;
	}
}


/**********
* Set requested object attribute. Specifying an unknown attribute will
* result in no changes.
*
* which_attr
*	Name of attribute to set.
*
* attr
*	The new attribute value.
**********/
void scene_object_t::set_attr_f(GLenum which_attr, GLfloat attr)
{
	switch (which_attr) {
	case OBJ_SPEED_X:
		move_speed[0] = attr;
		break;
	case OBJ_SPEED_Y:
		move_speed[1] = attr;
		break;
	case OBJ_SPEED_Z:
		move_speed[2] = attr;
		break;
	case OBJ_SPEED_PITCH:
		rotate_speed[0] = attr;
		break;
	case OBJ_SPEED_YAW:
		rotate_speed[1] = attr;
		break;
	case OBJ_SPEED_ROLL:
		rotate_speed[2] = attr;
		break;
	case OBJ_MAX_MOVE_SPEED:
		max_move_speed = attr;
		break;
	case OBJ_MAX_ROTATE_SPEED:
		max_rotate_speed = attr;
		break;
	case OBJ_FRONT_ANGLE:
		glRotatef(-front_angle, 1, 0, 0);
		front_angle = attr;
		glRotatef(front_angle, 1, 0, 0);
		break;
	default:
		break;
	}
}


/**********
* Set requested object attribute. Specifying an unknown attribute will
* result in no changes.
*
* which_attr
*	Name of attribute to set.
*
* attr
*	The new attribute value. Thee attribute specified by which_attr is
*	expected to be a vector or a set of coordinates, so attr must be a
*	3-element array to contain the full attribute.
**********/
void scene_object_t::set_attr_fv(GLenum which_attr, GLfloat *attr)
{
	switch (which_attr) {
	case OBJ_FRONT:
		break;
	default:
		break;
	}
}


/**********
* Places the object at specified location facing the specified point,
* with the up direction as specified.
*
* front_angle is reset to 0.
**********/
void scene_object_t::place_at(GLfloat new_pos_x, GLfloat new_pos_y,
		GLfloat new_pos_z, GLfloat lookat_x, GLfloat lookat_y,
		GLfloat lookat_z, GLfloat up_x, GLfloat up_y, GLfloat up_z)
{
	GLfloat m[16];

	glPushMatrix();
	glLoadIdentity();
	gluLookAt(new_pos_x, new_pos_y, new_pos_z,
			lookat_x, lookat_y, lookat_z,
			up_x, up_y, up_z);
	glGetFloatv(GL_MODELVIEW_MATRIX, m);
	invert_matrix(m, transformation);
	glPopMatrix();
	front_angle = 0;
}


/**********
* Render the object, taking into account its position and orientation.
*
* Relies on OpenGL's internal clipping and culling.
**********/
void scene_object_t::draw(void)
{
	glPushMatrix();
	glMultMatrixf(transformation);
	draw_here();
	glPopMatrix();
}


/**********
* Render the object, ignoring position and orientation. I.e. draw the
* object wherever the current (MODELVIEW) matrix dictates.
*
* Derived classes should override this function.
**********/
void scene_object_t::draw_here()
{
}


/**********
* Initializes object with specified attributes.
**********/
scene_object_t::scene_object_t(GLfloat init_pos_x,
		GLfloat init_pos_y,
		GLfloat init_pos_z,
		GLfloat lookat_x,
		GLfloat lookat_y,
		GLfloat lookat_z,
		GLfloat up_x,
		GLfloat up_y,
		GLfloat up_z)
{
	place_at(init_pos_x, init_pos_y, init_pos_z,
			lookat_x, lookat_y, lookat_z,
			up_x, up_y, up_z);

	max_move_speed = 0;
	max_rotate_speed = 0;
	max_durability = INFINITY;
	move_speed[0] = move_speed[1] = move_speed[2] = 0;
	rotate_speed[0] = rotate_speed[1] = rotate_speed[2] = 0;
	durability = max_durability;

	use_inertial = 1;
}