/* Base class for a solid object within a 3D scene.
 *
 * GL rendering context must be initialized prior to creating an
 * instance of this object.
 *
 * For most derived objects, draw_here() will usually be overridden.
 */


#ifndef _SCENE_OBJECT_H
#define _SCENE_OBJECT_H


#include <GL/glut.h>
#include "more_math.h"

// Bit flags to indicate whether the object is moving (or rotating) in
// a certain direction.
/* Placeholder */

// get_attr() and set_attr() parameters used to specify the object's
// attributes.
#define OBJ_FRONT		0	// Unit vector to indicate object's front (relative to parent object).
#define OBJ_DURABILITY	1	// Object's durability.
#define OBJ_SPEED_X		2	// Object's speed along local x-axis.
#define OBJ_SPEED_Y		3
#define OBJ_SPEED_Z		4
#define OBJ_SPEED_PITCH	5	// Object's speed about local x-axis.
#define OBJ_SPEED_YAW	6
#define OBJ_SPEED_ROLL	7
#define OBJ_MAX_MOVE_SPEED	8	// Not currently used.
#define OBJ_MAX_ROTATE_SPEED	9	// Not currently used.
#define OBJ_MAX_DURABILITY	10
#define OBJ_USE_INERTIAL	11	// Use intertial axes for movement/rotation.
#define OBJ_FRONT_ANGLE	12	// Angle between body's front and inertial front.

class scene_object_t {
public:
	void move_back(void);
	void move_forward(void) { move_speed[2] = -move_speed[2]; move_back(); move_speed[2] = -move_speed[2]; }
	void move_right(void);
	void move_left(void) { move_speed[0] = -move_speed[0]; move_right(); move_speed[0] = -move_speed[0]; }
	void move_up(void);
	void move_down(void) { move_speed[1] = -move_speed[1]; move_up(); move_speed[1] = -move_speed[1]; }
	void pitch_back(GLfloat angle);
	void pitch_back(void);
	void pitch_forward(GLfloat angle) { rotate_speed[0] = -rotate_speed[0]; pitch_back(angle); rotate_speed[0] = -rotate_speed[0]; }
	void pitch_forward(void) { rotate_speed[0] = -rotate_speed[0]; pitch_back(); rotate_speed[0] = -rotate_speed[0]; }
	void yaw_left(GLfloat angle);
	void yaw_left(void);
	void yaw_right(GLfloat angle) { rotate_speed[1] = -rotate_speed[1]; yaw_left(angle); rotate_speed[1] = -rotate_speed[1]; }
	void yaw_right(void) { rotate_speed[1] = -rotate_speed[1]; yaw_left(); rotate_speed[1] = -rotate_speed[1]; }
	void roll_left(GLfloat angle);
	void roll_left(void);
	void roll_right(GLfloat angle) { rotate_speed[2] = -rotate_speed[2]; roll_left(angle); rotate_speed[2] = -rotate_speed[2]; }
	void roll_right(void) { rotate_speed[2] = -rotate_speed[2]; roll_left(); rotate_speed[2] = -rotate_speed[2]; }
	GLint get_attr_i(GLenum which_attr);
	GLfloat get_attr_f(GLenum which_attr);
	void get_attr_fv(GLenum which_attr, GLfloat *attr);
	void set_attr_i(GLenum which_attr, GLint attr);
	void set_attr_f(GLenum which_attr, GLfloat attr);
	void set_attr_fv(GLenum which_attr, GLfloat *attr);
	void place_at(GLfloat new_pos_x, GLfloat new_pos_y, GLfloat new_pos_z,
			GLfloat lookat_x, GLfloat lookat_y, GLfloat lookat_z,
			GLfloat up_x, GLfloat up_y, GLfloat up_z);
	void draw(void);
	scene_object_t(GLfloat init_pos_x = 0.0f,	// Coordinate.
			GLfloat init_pos_y = 0.0f,
			GLfloat init_pos_z = 0.0f,
			GLfloat lookat_x = 0.0f,	// Coordinate.
			GLfloat lookat_y = 0.0f,
			GLfloat lookat_z = -1.0f,
			GLfloat up_x = 0.0f,	// Vector.
			GLfloat up_y = 1.0f,
			GLfloat up_z = 0.0f);

protected:
	virtual void draw_here(void);
	GLfloat scale_factor(void);
	GLfloat	move_speed[3];	// Object's desired movement speed along each axis.
	GLfloat	rotate_speed[3];	// Desired rotational speed about each axis.
	GLint	durability;	// Object's durability.
	GLfloat	max_move_speed;	// Not currently used.
	GLfloat	max_rotate_speed;	// Not currently used.
	GLint	max_durability;
	GLint	use_inertial;	// Whether to use intertial axes for movement/rotation.

	GLfloat transformation[16];	// Object's transformation matrix.
	GLfloat	front_angle;	// Angle between body's front and intertial front.

	// See character animation FAQ.
	GLint	current_frameset;	// Which set of animation frames is current.
};


#endif